from mloggers.logger import Logger


class OptionalLogger(Logger):
    """
    A wrapper for a logger which can be `None`.
    This object can be useful library-side to not force your users to use `mloggers`.
    The benefit of this wrapper is that you never have to use `if logger is not None:` again!

    Example:
    ```python
    from mloggers import Logger, OptionalLogger

    def some_function(logger: Logger | None):
        my_logger = OptionalLogger(logger)

        # If the logger is None, nothing will happen (not even an error)
        my_logger.info("This will only log if the logger is not None.")
    ```
    """

    def __init__(self, logger: Logger | None = None):
        """
        Initialize the OptionalLogger.

        ### Parameters
        - [optional] `logger`: the logger to wrap.
        """

        # Hopefully this name is unique enough
        self.__internal_wrapped_logger__ = logger

    def log(self, _):
        pass

    def __getattribute__(self, attr: str):
        """Re-route all attribute access to the logger if it exists."""

        if attr == "__internal_wrapped_logger__":
            return object.__getattribute__(self, attr)

        try:
            return getattr(self.__internal_wrapped_logger__, attr)
        except AttributeError:
            return lambda *_: None



from abc import ABC, abstractmethod
from typing import Any, Callable

import numpy.typing as npt

from mloggers._log_levels import LogLevel, _log_level_properties

# This constant is used to assign an importance level to anything not using the LogLevel enum.
# It was chosen to be the same as `LogLevel.INFO`, but it can be changed to any other value.
DEFAULT_IMPORTANCE = _log_level_properties[LogLevel.INFO].priority  # type:ignore[reportAttributeAccessIssue]


class Logger(ABC):
    """The abstract class for a logger."""

    def __init__(self, default_priority: LogLevel | int = LogLevel.INFO):  # type:ignore[reportArgumentType]
        """
        Initialize the logger.

        ### Parameters
        ----------
        `default_priority`: The default log level priority to use.
        - This parameter filters out messages with a lower importance level than the one provided. It can be either a `LogLevel` object or an integer.
        - When calling the logger with a level not from the `LogLevel` enum, the importance level will be set to 0 (same as `LogLevel.INFO`).
        - For example, if the log level is set to `LogLevel.INFO`, only messages with a level of `LogLevel.INFO` or higher will be printed (which excludes `LogLevel.DEBUG`).
        """

        self._min_priority = (
            _log_level_properties[default_priority].priority
            if isinstance(default_priority, LogLevel)
            else default_priority
        )

    @abstractmethod
    def log(
        self,
        *messages: str | dict[str, Any] | list[Any] | npt.NDArray[Any],
        level: LogLevel | str | None = None,
        **kwargs: Any,
    ):
        """
        Log a message.

        ### Parameters
        ----------
        `messages`: the messages to log.
        - These can be any number of messages, separated by commas. They can be of the following types:
            - If a stringifiable object (implements `__str__()`), the message will be logged as-is.
            - If a list or numpy array, the message will be logged as a stringified list.
            - If a dictionary, the message will be logged as a JSON string.
                - The dictionary must be JSON serializable.
                - You can provide None dictionary values to mean that the key is a header or title of the message.
        - If multiple messages are provided, they must be either all dictionaries or none of them. Strings, lists and arrays will be joined into a single string, while dictionaries will be printed as separate log entries.
        `level`: the level of the message (e.g., INFO, WARN, ERROR, DEBUG, etc.).
        - If None, no level will be printed.
        - If a string is provided, it will be colored in green (when colors are used) and uppercased; otherwise, the color will be the one associated with the `LogLevel` at time of registration.

        ### Raises
        ----------
        `TypeError`: if the message is not a string, a dictionary or does not implement `__str__()`.
        `TypeError`: if the messages are a mix of strings and dictionaries.
        """

        # Check if the messages are of the correct type.
        for message in messages:
            if (
                not isinstance(message, dict)
                and not hasattr(message, "__str__")
                and not callable(getattr(message, "__str__"))
            ):
                raise TypeError(
                    f"Expected message to be a string, a dictionary or to have implemented __str__(), but got {type(message)}."
                )

        # Check if there is both a string and a dictionary in the messages.
        if any(isinstance(message, dict) for message in messages) and any(
            not isinstance(message, dict) for message in messages
        ):
            raise TypeError(
                "Expected all messages to be either strings or dictionaries, but got a mix of both."
            )

        # Filter out messages with a lower importance level than the current priority.
        if (
            isinstance(level, LogLevel)
            and _log_level_properties[level].priority < self._min_priority
        ):
            return False
        elif isinstance(level, str) and DEFAULT_IMPORTANCE < self._min_priority:
            return False
        return True

    def set_min_priority(self, value: LogLevel | int):
        """
        Set the minimum log level priority.

        ### Parameters
        ----------
        `value`: the log level priority to set.
        - If a `LogLevel`, the priority of that log level will be considered.
        - If an `int`, such number will be considered.
        """

        if isinstance(value, int):
            self._min_priority = value
        else:
            self._min_priority = _log_level_properties[value].priority

    def _call_impl(self, *args, **kwargs):
        return self.log(*args, **kwargs)

    __call__: Callable[..., Any] = _call_impl

    def info(
        self,
        *messages: str | dict[str, Any] | list[Any] | npt.NDArray[Any],
        **kwargs: Any,
    ):
        """
        Wrapper for calling `log` with `level=LogLevel.INFO`.

        ### Parameters
        ----------
        `messages`: the messages to log.
        - These can be any number of messages, separated by commas. They can be of the following types:
            - If a stringifiable object (implements `__str__()`), the message will be logged as-is.
            - If a list or numpy array, the message will be logged as a stringified list.
            - If a dictionary, the message will be logged as a JSON string.
                - The dictionary must be JSON serializable.
                - You can provide None dictionary values to mean that the key is a header or title of the message.
        - If multiple messages are provided, they must be either all dictionaries or none of them. Strings, lists and arrays will be joined into a single string, while dictionaries will be printed as separate log entries.
        """

        self.log(*messages, level=LogLevel.INFO, **kwargs)  # type:ignore[reportArgumentType]

    def warn(
        self,
        *messages: str | dict[str, Any] | list[Any] | npt.NDArray[Any],
        **kwargs: Any,
    ):
        """
        Wrapper for calling `log` with `level=LogLevel.WARN`.

        ### Parameters
        ----------
        `messages`: the messages to log.
        - These can be any number of messages, separated by commas. They can be of the following types:
            - If a stringifiable object (implements `__str__()`), the message will be logged as-is.
            - If a list or numpy array, the message will be logged as a stringified list.
            - If a dictionary, the message will be logged as a JSON string.
                - The dictionary must be JSON serializable.
                - You can provide None dictionary values to mean that the key is a header or title of the message.
        - If multiple messages are provided, they must be either all dictionaries or none of them. Strings, lists and arrays will be joined into a single string, while dictionaries will be printed as separate log entries.
        """

        self.log(*messages, level=LogLevel.WARN, **kwargs)  # type:ignore[reportArgumentType]

    # Alias warning to warn
    warning = warn

    def error(
        self,
        *messages: str | dict[str, Any] | list[Any] | npt.NDArray[Any],
        **kwargs: Any,
    ):
        """
        Wrapper for calling `log` with `level=LogLevel.ERROR`.

        ### Parameters
        ----------
        `messages`: the messages to log.
        - These can be any number of messages, separated by commas. They can be of the following types:
            - If a stringifiable object (implements `__str__()`), the message will be logged as-is.
            - If a list or numpy array, the message will be logged as a stringified list.
            - If a dictionary, the message will be logged as a JSON string.
                - The dictionary must be JSON serializable.
                - You can provide None dictionary values to mean that the key is a header or title of the message.
        - If multiple messages are provided, they must be either all dictionaries or none of them. Strings, lists and arrays will be joined into a single string, while dictionaries will be printed as separate log entries.
        """

        self.log(*messages, level=LogLevel.ERROR, **kwargs)  # type:ignore[reportArgumentType]

    def debug(
        self,
        *messages: str | dict[str, Any] | list[Any] | npt.NDArray[Any],
        **kwargs: Any,
    ):
        """
        Wrapper for calling `log` with `level=LogLevel.DEBUG`.

        ### Parameters
        ----------
        `messages`: the messages to log.
        - These can be any number of messages, separated by commas. They can be of the following types:
            - If a stringifiable object (implements `__str__()`), the message will be logged as-is.
            - If a list or numpy array, the message will be logged as a stringified list.
            - If a dictionary, the message will be logged as a JSON string.
                - The dictionary must be JSON serializable.
                - You can provide None dictionary values to mean that the key is a header or title of the message.
        - If multiple messages are provided, they must be either all dictionaries or none of them. Strings, lists and arrays will be joined into a single string, while dictionaries will be printed as separate log entries.
        """

        self.log(*messages, level=LogLevel.DEBUG, **kwargs)  # type:ignore[reportArgumentType]



"""
A collection of loggers well-suited for machine learning experiments.
"""

from ._log_levels import LogLevel, register_level
from .console_logger import ConsoleLogger
from .file_logger import FileLogger
from .logger import Logger
from .multi_logger import MultiLogger
from .optional_logger import OptionalLogger
from .wandb_logger import WandbLogger

__all__ = [
    "ConsoleLogger",
    "FileLogger",
    "LogLevel",
    "Logger",
    "MultiLogger",
    "OptionalLogger",
    "WandbLogger",
    "register_level",
]



from typing import Any

import numpy.typing as npt

from mloggers._log_levels import LogLevel
from mloggers.logger import Logger


class MultiLogger(Logger):
    """Logs to multiple loggers."""

    def __init__(
        self,
        loggers: list[Logger],
        default_mask: list[type[Logger]] = [],
        default_priority: LogLevel | int = LogLevel.INFO,  # type:ignore[reportArgumentType]
    ):
        """
        Initializes a multi-logger.

        ### Parameters
        ----------
        `loggers`: a list of the initialized loggers to use.
        `default_mask`: the default mask to use when logging.
        `default_priority`: The default log level priority to use.
        """

        super().__init__(default_priority)

        self._loggers = loggers
        self._default_mask = default_mask

        for logger in self._loggers:
            logger.set_min_priority(self._min_priority)

    def set_min_priority(self, level: LogLevel | int):
        super(MultiLogger, self).set_min_priority(level)

        for logger in self._loggers:
            logger.set_min_priority(level)

    def log(
        self,
        *messages: str | dict[str, Any] | list[Any] | npt.NDArray[Any],
        level: LogLevel | str | None = None,
        mask: list[type[Logger]] | None = None,
        **kwargs: Any,
    ):
        """
        Logs a message to multiple loggers.

        ### Parameters
        ----------
        `messages`: the messages to log.
        - These can be any number of messages, separated by commas. They can be of the following types:
            - If a stringifiable object (implements `__str__()`), the message will be logged as-is.
            - If a list or numpy array, the message will be logged as a stringified list.
            - If a dictionary, the message will be logged as a JSON string.
                - The dictionary must be JSON serializable.
                - You can provide None dictionary values to mean that the key is a header or title of the message.
        - If multiple messages are provided, they must be either all dictionaries or none of them. Strings, lists and arrays will be joined into a single string, while dictionaries will be printed as separate log entries.
        `level`: the level of the message (e.g., INFO, WARN, ERROR, DEBUG, etc.).
        - If None, no level will be printed.
        - If a string is provided, it will be colored in green (when colors are used) and uppercased; otherwise, the color will be the one associated with the `LogLevel` at time of registration.

        ### Raises
        ----------
        `TypeError`: if the message is not a string, a dictionary, list, numpy array or does not implement `__str__()`.
        `TypeError`: if the messages are a mix of strings and dictionaries.
        """

        # NOTE: No need for checking the validity of the message, as the individual loggers will do that.
        # super(MultiLogger, self).log(message, level, *args, **kwargs)

        if mask is None:
            mask = self._default_mask

        for logger in [
            logger
            for logger in self._loggers
            if not any(isinstance(logger, type) for type in mask)
        ]:
            logger(*messages, level=level, **kwargs)

    def info(
        self,
        *messages: str | dict[str, Any] | list[Any] | npt.NDArray[Any],
        mask: list[type[Logger]] | None = None,
        **kwargs: Any,
    ):
        """
        Wrapper for calling `log` with `level=LogLevel.INFO`.

        ### Parameters
        ----------
        `messages`: the messages to log.
        - These can be any number of messages, separated by commas. They can be of the following types:
            - If a stringifiable object (implements `__str__()`), the message will be logged as-is.
            - If a list or numpy array, the message will be logged as a stringified list.
            - If a dictionary, the message will be logged as a JSON string.
                - The dictionary must be JSON serializable.
                - You can provide None dictionary values to mean that the key is a header or title of the message.
        - If multiple messages are provided, they must be either all dictionaries or none of them. Strings, lists and arrays will be joined into a single string, while dictionaries will be printed as separate log entries.
        `mask`: a list of logger names to not be used to log this message.
        - If None, the default mask will be used.
        """

        self.log(*messages, level=LogLevel.INFO, mask=mask, **kwargs)  # type:ignore[reportArgumentType]

    def warn(
        self,
        *messages: str | dict[str, Any] | list[Any] | npt.NDArray[Any],
        mask: list[type[Logger]] | None = None,
        **kwargs: Any,
    ):
        """
        Wrapper for calling `log` with `level=LogLevel.WARN`.

        ### Parameters
        ----------
        `messages`: the messages to log.
        - These can be any number of messages, separated by commas. They can be of the following types:
            - If a stringifiable object (implements `__str__()`), the message will be logged as-is.
            - If a list or numpy array, the message will be logged as a stringified list.
            - If a dictionary, the message will be logged as a JSON string.
                - The dictionary must be JSON serializable.
                - You can provide None dictionary values to mean that the key is a header or title of the message.
        - If multiple messages are provided, they must be either all dictionaries or none of them. Strings, lists and arrays will be joined into a single string, while dictionaries will be printed as separate log entries.
        `mask`: a list of logger names to not be used to log this message.
        - If None, the default mask will be used.
        """

        self.log(*messages, level=LogLevel.WARN, mask=mask, **kwargs)  # type:ignore[reportArgumentType]

    # Alias warning to warn
    warning = warn

    def error(
        self,
        *messages: str | dict[str, Any] | list[Any] | npt.NDArray[Any],
        mask: list[type[Logger]] | None = None,
        **kwargs: Any,
    ):
        """
        Wrapper for calling `log` with `level=LogLevel.ERROR`.

        ### Parameters
        ----------
        `messages`: the messages to log.
        - These can be any number of messages, separated by commas. They can be of the following types:
            - If a stringifiable object (implements `__str__()`), the message will be logged as-is.
            - If a list or numpy array, the message will be logged as a stringified list.
            - If a dictionary, the message will be logged as a JSON string.
                - The dictionary must be JSON serializable.
                - You can provide None dictionary values to mean that the key is a header or title of the message.
        - If multiple messages are provided, they must be either all dictionaries or none of them. Strings, lists and arrays will be joined into a single string, while dictionaries will be printed as separate log entries.
        `mask`: a list of logger names to not be used to log this message.
        - If None, the default mask will be used.
        """

        self.log(*messages, level=LogLevel.ERROR, mask=mask, **kwargs)  # type:ignore[reportArgumentType]

    def debug(
        self,
        *messages: str | dict[str, Any] | list[Any] | npt.NDArray[Any],
        mask: list[type[Logger]] | None = None,
        **kwargs: Any,
    ):
        """
        Wrapper for calling `log` with `level=LogLevel.DEBUG`.

        ### Parameters
        ----------
        `messages`: the messages to log.
        - These can be any number of messages, separated by commas. They can be of the following types:
            - If a stringifiable object (implements `__str__()`), the message will be logged as-is.
            - If a list or numpy array, the message will be logged as a stringified list.
            - If a dictionary, the message will be logged as a JSON string.
                - The dictionary must be JSON serializable.
                - You can provide None dictionary values to mean that the key is a header or title of the message.
        - If multiple messages are provided, they must be either all dictionaries or none of them. Strings, lists and arrays will be joined into a single string, while dictionaries will be printed as separate log entries.
        `mask`: a list of logger names to not be used to log this message.
        - If None, the default mask will be used.
        """

        self.log(*messages, level=LogLevel.DEBUG, mask=mask, **kwargs)  # type:ignore[reportArgumentType]



from datetime import datetime
from typing import Any

import numpy.typing as npt
from termcolor import colored

from mloggers._log_levels import LogLevel, _log_level_properties
from mloggers.logger import Logger
from mloggers.utils import serialize


class ConsoleLogger(Logger):
    """Logs to the console (i.e., standard I/O)."""

    def __init__(self, default_priority: LogLevel = LogLevel.INFO):  # type:ignore[reportArgumentType]
        super().__init__(default_priority)

    def log(
        self,
        *messages: str | dict[str, Any] | list[Any] | npt.NDArray[Any],
        level: LogLevel | str | None = None,
        **kwargs: Any,
    ):
        if not super(ConsoleLogger, self).log(*messages, level=level, **kwargs):
            return

        time = "[" + datetime.now().strftime("%H:%M:%S") + "]"

        if level is None:
            level_str = ""
            level_clr = ""
        else:
            # Create the level string to be printed
            if isinstance(level, LogLevel):
                level_str = "[" + level.name + "] "
            else:
                level_str = "[" + str(level).upper() + "] "

            # Color the level string
            if isinstance(level, LogLevel):
                level_clr = colored(level_str, _log_level_properties[level].color)  # type:ignore[reportArgumentType]
            else:
                level_clr = colored(level_str, "green")

        # The first level of the dictionary is printed as a multiline indented message.
        # The rest of the levels are printed as a single line prettifyed depending on the type of the value.

        # Handle multiple messages
        if len(messages) > 1:
            messages = list(messages)
            # If the messages are strings, join them into a single string
            if all(
                hasattr(message, "__str__")
                and callable(getattr(message, "__str__"))
                and not isinstance(message, dict)
                for message in messages
            ):
                messages = " ".join([str(message) for message in messages])
            # If the messages are dictionaries, log them separately
            else:
                for message in messages:
                    self.log(message, level=level)
                return
        else:
            messages = messages[0]

        if isinstance(messages, dict):
            first = True
            for key, value in messages.items():
                if not first:
                    time = " " * len(time)
                    level_clr = " " * len(level_str)

                if isinstance(value, float):
                    print(f"{level_clr}{time} {key}: {value:.5f}")
                elif value is None:  # Used for headers, titles, etc.
                    print(f"{level_clr}{time} {key}")
                else:
                    try:
                        value = serialize(value)
                        print(f"{level_clr}{time} {key}: {value}")
                    except TypeError as e:
                        print(
                            f'{colored("[ERROR]", "red")} [ConsoleLogger] Could not convert the message to a JSON serializable format: {e}'
                        )
                        continue

                first = False

        elif hasattr(messages, "__str__") and callable(getattr(messages, "__str__")):
            print(f"{level_clr}{time} {str(messages)}")

        else:
            try:
                serialized = serialize(messages)
                print(f"{level_clr}{time} {serialized}")
            except TypeError as e:
                print(
                    f'{colored("[ERROR]", "red")} [ConsoleLogger] Could not convert the message to a JSON serializable format: {e}'
                )



import json
import os
from datetime import datetime
from typing import Any

import numpy.typing as npt
from termcolor import colored

from mloggers._log_levels import LogLevel
from mloggers.logger import Logger
from mloggers.utils import serialize


class FileLogger(Logger):
    """Logs to a file."""

    def __init__(
        self,
        file_path: str,
        default_priority: LogLevel | int = LogLevel.INFO,  # type:ignore[reportArgumentType]
    ):
        """
        Initializes a file logger.

        ### Parameters
        ----------
        `file_path`: the path to the file to log to.
        - The file will be created if it does not exist. If it does, the logs will be appended to it.
        `default_priority`: The default log level priority to use.
        """

        super().__init__(default_priority)

        # Create the file if it does not exist
        if not os.path.exists(file_path):
            dir_path = os.path.dirname(file_path)
            try:
                # If dir path is empty, it means the file is in the current directory
                if dir_path != "":
                    os.makedirs(dir_path)
            except FileExistsError:
                pass
            with open(file_path, "w") as file:
                file.write("")

        print(f'{colored("[INFO]", "cyan")} [FileLogger] Logging to file {file_path}')

        self._file_path = file_path

    def log(
        self,
        *messages: str | dict[str, Any] | list[Any] | npt.NDArray[Any],
        level: LogLevel | str | None = None,
        **kwargs: Any,
    ):
        if not super(FileLogger, self).log(*messages, level=level, **kwargs):
            return

        # Handle multiple messages
        if len(messages) > 1:
            messages = list(messages)
            # If the messages are strings, join them into a single string.
            if all(
                hasattr(message, "__str__")
                and callable(getattr(message, "__str__"))
                and not isinstance(message, dict)
                for message in messages
            ):
                message = " ".join([str(message) for message in messages])
            # If the messages are dictionaries, log them separately.
            else:
                for message in messages:
                    self.log(message, level=level)
                return
        else:
            message = messages[0]

        # JSON-serialize the message
        try:
            serialized_message = serialize(message)
        except TypeError as e:
            print(
                f'{colored("[ERROR]", "red")} [FileLogger] Could not convert the message to a JSON serializable format: {e}'
            )
            return

        # Read the existing logs
        try:
            with open(self._file_path, "r") as file:
                existing_content = file.read()
                if (
                    existing_content == ""
                    or existing_content.isspace()
                    or existing_content == "[]"
                ):
                    prev_logs = []
                else:
                    try:
                        file.seek(0)
                        prev_logs = json.load(file)
                    except json.decoder.JSONDecodeError:
                        print(
                            f'{colored("[WARNING]", "yellow")} [FileLogger] Could not decode existing logs, new logs will not be appended to the file.'
                        )
                        return
        except Exception as e:
            print(
                f'{colored("[ERROR]", "red")} [FileLogger] Exception thrown while reading from the logging file: {e}'
            )
            return

        # Create the new log and write it to the file
        new_logs = prev_logs.copy()
        try:
            log: dict[str, Any] = {
                "timestamp": datetime.now().strftime("%d/%m/%Y %H:%M:%S"),
            }
            if level is not None:
                log["level"] = (
                    level.name if isinstance(level, LogLevel) else str(level).upper()
                )
            log["message"] = serialized_message
            new_logs.append(log)

            with open(self._file_path, "w") as file:
                file.seek(0)
                json.dump(new_logs, file, indent=4)

        except Exception as e:
            print(
                f'{colored("[ERROR]", "red")} [FileLogger] Exception thrown while logging to a file: {e}'
            )

            # Rollback the changes
            try:
                with open(self._file_path, "w") as file:
                    file.seek(0)
                    json.dump(prev_logs, file, indent=4)
            except Exception as e:
                print(
                    f'{colored("[ERROR]", "red")} [FileLogger] Exception thrown while rolling back the changes: {e}'
                )



import sys
from dataclasses import dataclass

from aenum import Enum, extend_enum


class LogLevel(Enum):  # type:ignore[reportGeneralTypeIssues]
    """
    The available log levels.
    Each level is associated with a color and an importance level.
    The importance level is used to set the minimum level of logs to display (all logs higher than the set level will be displayed).
    To register a new level use `mloggers.register_level`.
    """

    @dataclass
    class Properties:
        color: str
        """The color to use when printing the log."""

        priority: int
        """The importance level of the log."""

    ERROR = "error"
    WARN = "warn"
    INFO = "info"
    DEBUG = "debug"


_log_level_properties: dict[str, LogLevel.Properties] = {
    LogLevel.ERROR: LogLevel.Properties(color="red", priority=sys.maxsize),
    LogLevel.WARN: LogLevel.Properties(color="yellow", priority=1),
    LogLevel.INFO: LogLevel.Properties(color="cyan", priority=0),
    LogLevel.DEBUG: LogLevel.Properties(color="magenta", priority=-1),
}


def register_level(name: str, properties: LogLevel.Properties):
    """
    Register a customized logger level, which will then be available as a member of `LogLevel`,
    where its `name` is the argument `level` and its `value` is the argument `level_info`.
    Note: the name of the level will be uppercased.

    ### Parameters
    ----------
    `level`: the level name to register.
    `properties`: the properties of the level.
    """

    extend_enum(LogLevel, name.upper(), name)
    _log_level_properties[name] = properties



from typing import Iterable, Sequence

from rich.progress import (
    BarColumn,
    MofNCompleteColumn,
    Progress,
    TextColumn,
    TimeElapsedColumn,
    TimeRemainingColumn,
)


# TODO: Improve, maybe with live display and support for nested progress bars
def log_progress(iterable: Iterable | Sequence):
    """
    Log an iterable or sequence using a progress bar.
    Wraps `rich.progress.Progress.track`.

    ### Parameters
    ----------
    `iterable`: the iterable whose progress to log.

    ### Returns
    -------
    The tracked iterable.
    """

    bar = Progress(
        TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
        BarColumn(),
        MofNCompleteColumn(),
        TextColumn("•"),
        TimeElapsedColumn(),
        TextColumn("•"),
        TimeRemainingColumn(),
        expand=True,
    )
    bar.__enter__()
    return bar.track(iterable)



from typing import Any

import numpy.typing as npt
import wandb
from omegaconf import DictConfig
from termcolor import colored

from mloggers._log_levels import LogLevel
from mloggers.logger import Logger


class WandbLogger(Logger):
    """Logs to Weights & Biases."""

    def __init__(
        self,
        project: str,
        group: str,
        experiment: str,
        default_priority: LogLevel | int = LogLevel.INFO,  # type:ignore[reportArgumentType]
        config: DictConfig | None = None,
    ):
        """
        Initializes a Weights & Biases logger.

        ### Parameters
        ----------
        `project`: the name of the project to log to.
        `group`: the name of the group to log to.
        `experiment`: the name of the experiment to log to.
        `default_priority`: the default log level priority to use.
        [optional] `config`: the configuration of the experiment.
        """

        super().__init__(default_priority)

        if config is not None:
            config = vars(config)
        wandb.init(project=project, group=group, name=experiment, config=config)

    def log(
        self,
        *messages: str | dict[str, Any] | list[Any] | npt.NDArray[Any],
        level: LogLevel | str | None = None,
        **kwargs: Any,
    ):
        if not super(WandbLogger, self).log(*messages, level=level, **kwargs):
            return

        # Handle multiple messages
        if len(messages) > 1:
            messages = list(messages)
            # If the messages are strings, join them into a single string.
            if all(
                hasattr(message, "__str__")
                and callable(getattr(message, "__str__"))
                and not isinstance(message, dict)
                for message in messages
            ):
                message = " ".join([str(message) for message in messages])
            # If the messages are dictionaries, log them separately.
            else:
                for message in messages:
                    self.log(message, level=level)
                return
        else:
            message = messages[0]

        if isinstance(message, dict):
            log = message
        else:
            message = str(
                message
            )  # This should be safe here as the super should have already checked that the message is a string or has a __str__ method.
            if level is not None:
                level_str = (
                    level.name if isinstance(level, LogLevel) else str(level).upper()
                )
                log = {level_str: message}
            else:
                log = {"message": message}

        try:
            wandb.log(log)
        except Exception as e:
            print(
                f'{colored("[ERROR]", "red")} [WandbLogger] Error while logging to wandb: {e}'
            )

    def __del__(self):
        """Finishes the connection with W&B."""

        wandb.finish()



from dataclasses import dataclass
from enum import Enum

from mloggers import ConsoleLogger, FileLogger, LogLevel


class MyEnum(Enum):
    A = 1
    B = 2
    C = 3


@dataclass
class SomeObject:
    a: int
    b: str


@dataclass
class SomeJSONObject:
    c: float
    d: bool

    def to_json(self):
        return {"c": self.c, "d": self.d}


# logger = FileLogger("test.log", default_priority=LogLevel.DEBUG)
logger = ConsoleLogger(default_priority=LogLevel.DEBUG)
logger.info("Hello, world!")
o1 = SomeObject(1, "hello")
o2 = SomeJSONObject(1.0, True)
logger.error("This is an error message.")
logger.error(2)
logger.error(2.532535)
logger.error(False)
logger.warn(o1)
logger.warn(o2)
logger.warn(MyEnum.B)
logger.debug(
    {
        "a": 1,
        "b": "hello",
        "object": o1,
        "json_object": o2,
    }
)
logger.debug(
    {
        "a": 1,
        "b": True,
        "c": 1.27453,
        "object": o1,
        "json_object": o2,
        "enum": MyEnum.A,
    }
)



import jsonpickle
import numpy as np


def serialize(message: object) -> object:
    """
    Serializes the message to a JSON serializable format.

    ### Parameters
    ----------
    `message`: The message to serialize.

    ### Returns
    ----------
    The serialized message.

    ### Raises
    ----------
    - `TypeError`: If the message cannot be serialized.
    """

    if isinstance(message, list):
        return [serialize(item) for item in message]
    if isinstance(message, np.ndarray):
        return [serialize(item) for item in message.tolist()]
    elif isinstance(message, dict):
        serialized_message = {}
        for key, value in message.items():
            serialized_message[key] = serialize(value)
        return serialized_message
    elif hasattr(message, "toJSON"):
        return message.toJSON()  # type:ignore[reportAttributeAccessIssue]
    elif hasattr(message, "to_json"):
        return message.to_json()  # type:ignore[reportAttributeAccessIssue]
    elif hasattr(message, "to_dict"):
        return message.to_dict()  # type:ignore[reportAttributeAccessIssue]
    elif hasattr(message, "__str__") and callable(getattr(message, "__str__")):
        return str(message)
    else:
        try:
            return dict(message)  # type:ignore[reportCallIssue, reportArgumentType]
        except Exception:
            try:
                return jsonpickle.encode(message)
            except Exception as e:
                raise TypeError(
                    f"Could not serialize the message: {message}. Error: {e}"
                )
